from textual.binding import Binding

VIM_OPTION_LIST_NAVIGATE = [
    Binding("j", "cursor_down", "Down", show=False),
    Binding("k", "cursor_up", "Up", show=False),
    Binding("G", "last", "Last", show=False),
    Binding("ctrl+b", "page_down", "Page Down", show=False),
    Binding("ctrl+f", "page_up", "Page Up", show=False),
]
BOOKMARK = Binding("b", "bookmark", "Bookmark")
CLEAR_CACHE = Binding("ctrl+d", "purge_from_cache", "Purge from cache")



from abc import ABC, abstractmethod
from typing import Literal


class Item(ABC):
    @property
    @abstractmethod
    def bookmarked(self) -> bool:
        raise NotImplementedError

    @bookmarked.setter
    @abstractmethod
    def bookmarked(self, value: bool) -> None:
        raise NotImplementedError

    @property
    @abstractmethod
    def cached(self) -> bool:
        raise NotImplementedError

    @property
    @abstractmethod
    def kind(self) -> Literal["resources", "providers"]:
        raise NotImplementedError

    @property
    @abstractmethod
    def display_name(self) -> str:
        raise NotImplementedError

    @property
    @abstractmethod
    def identifying_name(self) -> str:
        raise NotImplementedError

    @abstractmethod
    def clear_from_cache(self) -> None:
        raise NotImplementedError



from tofuref.widgets.code_block_select import CodeBlockSelect
from tofuref.widgets.content_window import ContentWindow
from tofuref.widgets.custom_rich_log import CustomRichLog
from tofuref.widgets.providers_option_list import ProvidersOptionList
from tofuref.widgets.resources_option_list import ResourcesOptionList
from tofuref.widgets.search_input import SearchInput

__all__ = [
    "CodeBlockSelect",
    "ContentWindow",
    "CustomRichLog",
    "ProvidersOptionList",
    "ResourcesOptionList",
    "SearchInput",
]



import os
import shutil
from pathlib import Path
from unittest.mock import patch

import pytest
from httpx import Response
from platformdirs import user_config_path

from tofuref.data.bookmarks import Bookmarks
from tofuref.data.cache import cached_file_path


@pytest.fixture(scope="session", autouse=True)
def clear_provider_index_cache(mock_cache_path: Path):
    cached_file = cached_file_path("index.json")
    cached_file.parent.mkdir(parents=True, exist_ok=True)
    if cached_file.exists():
        cached_file.unlink()
    fallback_file = Path(__file__).parent.parent / "tofuref" / "fallback" / "providers.json"
    shutil.copy(str(fallback_file), str(cached_file))
    print(str(fallback_file))
    yield
    if cached_file.exists():
        cached_file.unlink()


@pytest.fixture(scope="session", autouse=True)
def config_file():
    """Yeah, let's add argparse for an alternative config file later, please"""
    config_file = user_config_path("tofuref") / "config.toml"
    backup_config_file = user_config_path("tofuref") / "config.toml.bak"
    moved = False
    if config_file.exists():
        moved = True
        shutil.move(str(config_file), str(backup_config_file))
    yield
    if moved:
        shutil.move(str(backup_config_file), str(config_file))


@pytest.fixture(scope="session")
def mock_cache_path():
    cache_dir = Path("__tests_cache__")

    def mock_user_cache_path(app_name, ensure_exists=False):
        if ensure_exists:
            cache_dir.mkdir(parents=True, exist_ok=True)
        return cache_dir

    with patch("tofuref.data.cache.user_cache_path", side_effect=mock_user_cache_path):
        yield cache_dir
    for file in cache_dir.glob("*"):
        file.unlink()
    cache_dir.rmdir()


@pytest.fixture
def clear_mock_cache(mock_cache_path: Path):
    for file in mock_cache_path.glob("*"):
        if file.name != "index.json":
            file.unlink()


@pytest.fixture(autouse=True)
def mock_http_requests():
    async def http_get(url, **kwargs):
        responses_dir = Path(__file__).parent / "responses"
        endpoint_map = {
            "https://pypi.org/pypi/tofuref/json": responses_dir / "pypi_tofuref.json",
            "https://api.opentofu.org/registry/docs/providers/hashicorp/aws/v6.0.0-beta1/index.json": responses_dir / "aws_600beta1_index.json",
            "https://api.opentofu.org/registry/docs/providers/hashicorp/aws/v6.0.0-beta1/index.md": responses_dir / "aws_600beta1_index.md",
            "https://api.opentofu.org/registry/docs/providers/integrations/github/v6.6.0/index.json": responses_dir / "github_660_index.json",
            "https://api.opentofu.org/registry/docs/providers/integrations/github/v6.6.0/index.md": responses_dir / "github_660_index.md",
            "https://api.opentofu.org/registry/docs/providers/integrations/github/v6.5.0/index.json": responses_dir / "github_650_index.json",
            "https://api.opentofu.org/registry/docs/providers/integrations/github/v6.5.0/index.md": responses_dir / "github_650_index.md",
            "https://api.opentofu.org/registry/docs/providers/integrations/github/v6.6.0/resources/actions_environment_secret.md": responses_dir
            / "github_action_env_secret.md",
            "https://api.opentofu.org/registry/docs/providers/integrations/github/v6.6.0/resources/membership.md": responses_dir
            / "github_membership.md",
        }

        return Response(200, content=endpoint_map[url].read_bytes())

    with patch("httpx.AsyncClient.get", side_effect=http_get) as mock_get:
        yield


@pytest.fixture(scope="session", autouse=True)
def disable_emoji():
    """Disabling emojis for tests, because they might look different depending on the OS, terminal etc."""
    os.environ["TOFUREF_THEME_EMOJI"] = "false"
    yield
    os.environ.pop("TOFUREF_THEME_EMOJI")


@pytest.fixture(scope="session", autouse=True)
def patch_bookmarks():
    class PatchedBookmarks(Bookmarks):
        def save_to_disk(self):
            pass

        def load_from_disk(self):
            self.saved = {"providers": [], "resources": []}

    with patch("tofuref.data.bookmarks.Bookmarks", PatchedBookmarks) as patched:
        yield patched



import json
import logging
from collections.abc import Collection
from pathlib import Path

from textual.widgets.option_list import Option

from tofuref.data.helpers import get_registry_api
from tofuref.data.providers import Provider
from tofuref.widgets.menu_option_list_base import MenuOptionListBase

LOGGER = logging.getLogger(__name__)


class ProvidersOptionList(MenuOptionListBase):
    def __init__(self, **kwargs):
        super().__init__(
            name="Providers",
            id="nav-provider",
            classes="nav-selector bordered",
            **kwargs,
        )
        self.border_title = "Providers"
        self.fallback_providers_file = Path(__file__).resolve().parent.parent / "fallback" / "providers.json"

    def populate(
        self,
        providers: Collection[Provider] | None = None,
    ) -> None:
        if providers is None:
            providers = self.app.providers.values()
        self.clear_options()
        self.border_subtitle = f"{len(providers)}/{len(self.app.providers)}"
        for provider in providers:
            self.add_option(provider)

    async def load_index(self) -> dict[str, Provider]:
        LOGGER.debug("Loading providers")
        providers = {}

        data = await get_registry_api(
            "index.json",
            log_widget=self.app.log_widget,
        )
        if not data:
            data = json.loads(self.fallback_providers_file.read_text())
            self.app.notify(
                "Something went wrong while fetching index of providers, using limited fallback.",
                title="Using fallback",
                severity="error",
            )

        LOGGER.debug("Got API response (or fallback)")

        for provider_json in data["providers"]:
            provider = Provider.from_json(provider_json)
            provider.log_widget = self.app.log_widget
            filter_in = (
                provider.versions,
                not provider.blocked,
                (not provider.fork_of or provider.organization == "opentofu"),
                provider.organization not in ["terraform-providers"],
            )
            if all(filter_in):
                providers[provider.display_name] = provider
                if self.app.bookmarks.check("providers", provider.identifying_name):
                    provider.bookmarked = True

        providers = {k: v for k, v in sorted(providers.items(), key=lambda p: (p[1].bookmarked, p[1].cached, p[1].popularity), reverse=True)}

        return providers

    async def on_option_selected(self, option: Option) -> None:
        provider_selected = option.prompt
        self.app.active_provider = provider_selected
        if self.app.fullscreen_mode:
            self.screen.maximize(self.app.navigation_resources)
        await self.app.navigation_resources.load_provider_resources(provider_selected)
        self.replace_option_prompt_at_index(self.highlighted, option.prompt)



from typing import ClassVar

from rich.console import Group
from rich.syntax import Syntax
from textual.binding import Binding, BindingType
from textual.widgets import OptionList
from textual.widgets.option_list import Option

from tofuref.config import config
from tofuref.widgets.keybindings import VIM_OPTION_LIST_NAVIGATE


class CodeBlockSelect(OptionList):
    BINDINGS: ClassVar[list[BindingType]] = [
        *OptionList.BINDINGS,
        Binding("escape", "close", "Close panel", show=False),
        *VIM_OPTION_LIST_NAVIGATE,
    ]

    def __init__(self, **kwargs):
        super().__init__(name="CodeBlocks", id="code-block-selector", classes="bordered", **kwargs)
        self.border_title = "Choose a codeblock to copy"

    def set_new_options(self, code_blocks: list[tuple]) -> None:
        self.clear_options()
        for i, (lexer, block) in enumerate(code_blocks):
            self.add_option(
                Group(
                    f"[b]Codeblock[/] {i + 1}",
                    Syntax(block, lexer=lexer if lexer else "hcl", theme=config.theme.codeblocks),
                )
            )
            self.add_option(None)
        self.focus()
        self.highlighted = 0

    def action_close(self):
        content_window = self.app.content_markdown
        content_window.document.focus()
        if not self.app.fullscreen_mode:
            self.screen.minimize()
        else:
            self.screen.maximize(content_window)
        self.parent.remove_children([self])

    async def on_option_selected(self, option: Option):
        code_selected = option.prompt.renderables[1].code
        self.app.copy_to_clipboard(code_selected)
        # We don't want the longest notification, so three dots will replace lines beyond the 3rd line
        code_selected_lines = code_selected.splitlines()
        lines_of_code_to_show = 4
        if len(code_selected_lines) > lines_of_code_to_show:
            snippet = "\n".join(code_selected_lines[:4])
            code_selected_notify = f"{snippet}\n..."
        else:
            code_selected_notify = code_selected.strip()
        self.app.notify(code_selected_notify, title="Copied to clipboard", markup=False)
        self.app.action_content()
        if not self.app.fullscreen_mode:
            self.screen.minimize()
        await self.parent.remove_children([self])



from typing import ClassVar

from textual.binding import BindingType
from textual.widgets import OptionList

from tofuref.data.meta import Item
from tofuref.widgets.keybindings import BOOKMARK, CLEAR_CACHE, VIM_OPTION_LIST_NAVIGATE


class MenuOptionListBase(OptionList):
    BINDINGS: ClassVar[list[BindingType]] = [*OptionList.BINDINGS, *VIM_OPTION_LIST_NAVIGATE, BOOKMARK, CLEAR_CACHE]

    async def action_bookmark(self):
        if self.highlighted is None:
            return
        option = self.get_option_at_index(self.highlighted)
        res: Item = option.prompt
        if not res.bookmarked:
            self.app.bookmarks.add(res.kind, res.identifying_name)
            res.bookmarked = True
            self.replace_option_prompt_at_index(self.highlighted, option.prompt)
            self.app.notify(f"{res.__class__.__name__} {res.display_name} bookmarked", title="Bookmark added")
        else:
            self.app.bookmarks.remove(res.kind, res.identifying_name)
            res.bookmarked = False
            self.replace_option_prompt_at_index(self.highlighted, option.prompt)
            self.app.notify(f"{res.__class__.__name__} {res.display_name} removed from bookmarks", title="Bookmark removed")

    async def action_purge_from_cache(self):
        if self.highlighted is None:
            return
        option = self.get_option_at_index(self.highlighted)
        res: Item = option.prompt
        res.clear_from_cache()
        self.replace_option_prompt_at_index(self.highlighted, option.prompt)
        self.app.notify(f"Provider {res.display_name} purged from cache", title="Cache purged")



from typing import ClassVar

from textual.binding import Binding, BindingType
from textual.widgets import Input


class SearchInput(Input):
    BINDINGS: ClassVar[list[BindingType]] = [
        *Input.BINDINGS,
        Binding("escape", "close", "Close panel", show=False),
    ]

    def __init__(self, **kwargs):
        super().__init__(placeholder="Search...", id="search", classes="bordered", **kwargs)
        self.border_title = "Search"

    def action_close(self):
        self.post_message(self.Changed(self, "", None))
        self.post_message(self.Submitted(self, "", None))



from dataclasses import dataclass
from enum import Enum
from typing import TYPE_CHECKING

from textual.content import Content

from tofuref.config import config
from tofuref.data import emojis
from tofuref.data.cache import cached_file_path, clear_from_cache
from tofuref.data.helpers import (
    get_registry_api,
    header_markdown_split,
)
from tofuref.data.meta import Item

if TYPE_CHECKING:
    from tofuref.data.providers import Provider


class ResourceType(Enum):
    RESOURCE = "resource"
    DATASOURCE = "datasource"
    GUIDE = "guide"
    FUNCTION = "function"


@dataclass
class Resource(Item):
    name: str
    provider: "Provider"
    type: ResourceType
    _content: str | None = None
    _cached: bool | None = None
    bookmarked: bool = False
    kind = "resources"

    def __lt__(self, other: "Resource") -> bool:
        return self.name < other.name

    def __gt__(self, other: "Resource") -> bool:
        return self.name > other.name

    @property
    def display_name(self):
        return self.name

    def visualize(self):
        cached_icon = emojis.CACHE if config.theme.emoji else "[$success]C[/] "
        bookmark_icon = emojis.BOOKMARK if config.theme.emoji else "[$success]B[/] "
        if self.bookmarked:
            prefix = bookmark_icon
        elif self.cached:
            prefix = cached_icon
        else:
            prefix = ""
        resource_icon = emojis.RESOURCE_TYPE[self.type.value] if config.theme.emoji else f"[$secondary]{self.type.value[0].upper()}[/]"
        return Content.from_markup(f"{resource_icon} {prefix}{self.name}")

    @property
    def identifying_name(self):
        return f"{self.provider.name}_{self.type.value}_{self.name}"

    def __hash__(self):
        return hash(self.identifying_name)

    @property
    def endpoint(self) -> str:
        return f"{self.provider.organization}/{self.provider.name}/{self.provider.active_version}/{self.type.value}s/{self.name}.md"

    @property
    def cached(self) -> bool:
        if self._cached is None:
            return cached_file_path(self.endpoint.replace(self.provider.active_version, "*"), glob=True).exists()
        return self._cached

    async def content(self):
        if self._content is None:
            doc_data = await get_registry_api(
                self.endpoint,
                json=False,
                log_widget=self.provider.log_widget,
            )
            _, self._content = header_markdown_split(doc_data)
            self._cached = True
        return self._content

    def clear_from_cache(self) -> None:
        if self.cached:
            clear_from_cache(self.endpoint)
            self._cached = False



from dataclasses import dataclass, field

from textual.constants import DEFAULT_THEME
from yaucl import BaseConfig, BaseSectionConfig


@dataclass
class ThemeConfig(BaseSectionConfig):
    ui: str = DEFAULT_THEME
    codeblocks: str = "material"
    borders_style: str = "ascii"
    emoji: bool = True


@dataclass
class Config(BaseConfig):
    """Can contain primitives or subclasses of ConfigSection"""

    theme: ThemeConfig = field(default_factory=ThemeConfig)
    http_request_timeout: float = 3.0
    index_cache_duration_days: int = 31
    fullscreen_init_threshold: int = 125


config = Config.init(app_name="tofuref")



import os
from unittest.mock import patch

import pytest
from packaging.version import Version

APP_PATH = "../tofuref/main.py"

SEARCH_GITHUB = ["s", "g", "i", "t", "h", "u", "b"]


def test_welcome(snap_compare, clear_mock_cache):
    assert snap_compare(APP_PATH, terminal_size=(200, 60))


def test_welcome_fullscreen(snap_compare, clear_mock_cache):
    assert snap_compare(APP_PATH)


@patch("tofuref.__version__", "1.0.0")
def test_welcome_update_available(snap_compare, clear_mock_cache):
    assert snap_compare(APP_PATH)


def test_toggle_fullscreen(snap_compare, clear_mock_cache):
    # Result: fullscreen mode off, even though it's a small window
    assert snap_compare(APP_PATH, press=["f"])


def test_content(snap_compare):
    assert snap_compare(APP_PATH, press=["c", "pagedown"])


def test_content_toc_on(snap_compare):
    assert snap_compare(APP_PATH, press=["c", "t"])


def test_content_toc_off(snap_compare):
    assert snap_compare(APP_PATH, press=["c", "t", "t"])


def test_content_toc_submit(snap_compare):
    assert snap_compare(APP_PATH, press=["c", "t", "down", "down", "down", "down", "down", "enter"])


def test_use_none_selected(snap_compare, clear_mock_cache):
    assert snap_compare(APP_PATH, press=["u"])


def test_search_github(snap_compare, clear_mock_cache):
    assert snap_compare(APP_PATH, press=SEARCH_GITHUB)


def test_search_github_cancel(snap_compare, clear_mock_cache):
    assert snap_compare(APP_PATH, press=[*SEARCH_GITHUB, "escape"])


def test_open_github(snap_compare):
    assert snap_compare(APP_PATH, press=[*SEARCH_GITHUB, "enter", "enter"])


def test_open_github_membership(snap_compare):
    assert snap_compare(
        APP_PATH,
        press=[*SEARCH_GITHUB, "enter", "enter", "s", "m", "e", "m", "b", "e", "r", "enter", "enter"],
    )


def test_back_to_providers(snap_compare, clear_mock_cache):
    assert snap_compare(APP_PATH, press=["enter", "p"])


def test_provider_overview(snap_compare):
    assert snap_compare(APP_PATH, press=[*SEARCH_GITHUB, "enter", "enter", "c"])


def test_version_picker(snap_compare, clear_mock_cache):
    assert snap_compare(APP_PATH, press=[*SEARCH_GITHUB, "enter", "enter", "v"])


def test_version_picker_submit(snap_compare):
    assert snap_compare(
        APP_PATH,
        press=[*SEARCH_GITHUB, "enter", "enter", "v", "down", "enter"],
    )


def test_use(snap_compare, clear_mock_cache):
    assert snap_compare(APP_PATH, press=[*SEARCH_GITHUB, "enter", "enter", "u"])


def test_copy_selection_github_overview(snap_compare):
    assert snap_compare(APP_PATH, press=[*SEARCH_GITHUB, "enter", "enter", "c", "y"])


def test_copy_selection_github_overview_copy_first(snap_compare):
    assert snap_compare(APP_PATH, press=[*SEARCH_GITHUB, "enter", "enter", "c", "y", "enter"])


def test_vim_providers(snap_compare, clear_mock_cache):
    assert snap_compare(APP_PATH, press=["j"])


def test_vim_resources(snap_compare, clear_mock_cache):
    assert snap_compare(APP_PATH, press=["enter", "j"])


def test_vim_content(snap_compare):
    assert snap_compare(APP_PATH, press=["c", "j"])


def test_config_theme(snap_compare):
    os.environ["TOFUREF_THEME_UI"] = "monokai"
    os.environ["TOFUREF_THEME_CODEBLOCKS"] = "monokai"
    os.environ["TOFUREF_THEME_BORDERS_STYLE"] = "solid"
    assert snap_compare(APP_PATH, press=[*SEARCH_GITHUB, "enter", "enter", "c", "wait:500", "pagedown"])
    os.environ.pop("TOFUREF_THEME_UI")
    os.environ.pop("TOFUREF_THEME_CODEBLOCKS")
    os.environ.pop("TOFUREF_THEME_BORDERS_STYLE")


def test_recent_provider(snap_compare, clear_mock_cache):
    assert snap_compare(APP_PATH, press=[*SEARCH_GITHUB, "enter", "enter", "p"])


def test_recent_resource(snap_compare, clear_mock_cache):
    assert snap_compare(APP_PATH, press=[*SEARCH_GITHUB, "enter", "enter", "enter", "r"])


def test_bookmark_provider(snap_compare):
    assert snap_compare(APP_PATH, press=[*SEARCH_GITHUB, "enter", "b"])


def test_bookmark_resource(snap_compare):
    assert snap_compare(APP_PATH, press=[*SEARCH_GITHUB, "enter", "enter", "b"])


def test_bookmark_cached_provider(snap_compare, clear_mock_cache):
    assert snap_compare(APP_PATH, press=[*SEARCH_GITHUB, "enter", "enter", "p", "b"])


def test_bookmark_cached_resource(snap_compare, clear_mock_cache):
    assert snap_compare(APP_PATH, press=[*SEARCH_GITHUB, "enter", "enter", "enter", "r", "b"])


def test_clear_from_cache(snap_compare, clear_mock_cache):
    assert snap_compare(APP_PATH, press=[*SEARCH_GITHUB, "enter", "enter", "p", "ctrl+d"])



from typing import cast

from textual.widgets.option_list import Option

from tofuref.data.providers import Provider
from tofuref.data.resources import Resource
from tofuref.widgets.menu_option_list_base import MenuOptionListBase


class ResourcesOptionList(MenuOptionListBase):
    def __init__(self, **kwargs):
        super().__init__(
            name="Resources",
            id="nav-resources",
            classes="nav-selector bordered",
            **kwargs,
        )
        self.border_title = "Resources"

    def populate(
        self,
        provider: Provider | None = None,
        resources: list[Resource] | None = None,
    ) -> None:
        self.clear_options()
        if provider is None:
            return
        self.border_subtitle = f"{provider.organization}/{provider.name} {provider.active_version}"

        if resources is None:
            for resource in provider.resources:
                self.add_option(resource)
        else:
            self.add_options(resources)

    async def load_provider_resources(
        self,
        provider: Provider,
    ):
        self.loading = True
        self.app.content_markdown.loading = True
        await provider.load_resources(bookmarks=self.app.bookmarks)
        self.app.content_markdown.update(await provider.overview())
        self.app.content_markdown.document.border_subtitle = f"{provider.display_name} {provider.active_version} Overview"
        self.populate(provider)
        self.focus()
        self.highlighted = 0
        self.app.content_markdown.loading = False
        self.loading = False

    async def on_option_selected(self, option: Option):
        resource_selected = cast(Resource, option.prompt)
        self.app.active_resource = resource_selected
        if self.app.fullscreen_mode:
            self.screen.maximize(self.app.content_markdown)
        self.app.content_markdown.loading = True
        was_cached = resource_selected.cached
        self.app.content_markdown.update(await resource_selected.content())
        is_cached = resource_selected.cached
        if was_cached != is_cached:
            self.replace_option_prompt_at_index(self.highlighted, option.prompt)
        self.app.content_markdown.document.border_subtitle = (
            f"{resource_selected.type.value} - {resource_selected.provider.name}_{resource_selected.name}"
        )
        self.app.content_markdown.document.focus()
        self.app.content_markdown.loading = False



import json

from tofuref.data.providers import Provider
from tofuref.widgets.providers_option_list import ProvidersOptionList

PROVIDER_JSON = json.loads("""{
  "addr": {
    "display": "0username/rabbitmq",
    "namespace": "0username",
    "name": "rabbitmq"
  },
  "link": "",
  "canonical_addr": null,
  "reverse_aliases": [],
  "description": "",
  "popularity": 0,
  "fork_count": 0,
  "fork_of": {
    "display": "",
    "namespace": "",
    "name": ""
  },
  "upstream_popularity": 0,
  "upstream_fork_count": 0,
  "versions": [
    {
      "id": "v1.9.1",
      "published": "2023-06-05T15:03:47Z"
    },
    {
      "id": "v1.8.1",
      "published": "2022-05-06T00:10:58Z"
    },
    {
      "id": "v1.7.0",
      "published": "2022-04-27T20:42:44Z"
    }
  ],
  "is_blocked": false
}
""")


def test_provider_use():
    provider = Provider.from_json(PROVIDER_JSON)
    assert (
        provider.use_configuration
        == """    rabbitmq = {
      source  = "0username/rabbitmq"
      version = "1.9.1"
    }"""
    )


def test_provider_fallback_exists():
    p = ProvidersOptionList()
    assert p.fallback_providers_file.exists()



import json as jsonlib
import logging
import re
from typing import Any

import httpx
from yaml import safe_load
from yaml.scanner import ScannerError

from tofuref import __version__
from tofuref.config import config
from tofuref.data.cache import cached_file_path, get_from_cache, save_to_cache

LOGGER = logging.getLogger(__name__)

CODEBLOCK_REGEX = r"^```([a-z]+)\n([\s\S]*?)^```"


def header_markdown_split(contents: str) -> tuple[dict, str]:
    """
    Most of the documentation files from the registry have a YAML "header"
    that we mostly (at the moment) don't care about. Either way we
    check for the header, and if it's there, we split it.
    """
    header = {}
    if re.match(r"^---$", contents, re.MULTILINE):
        split_contents = re.split(r"^---$", contents, maxsplit=2, flags=re.MULTILINE)
        try:
            header = safe_load(split_contents[1])
        except ScannerError as _:
            header = {}
        markdown_content = split_contents[2]
    else:
        markdown_content = contents
    return header, markdown_content


async def get_registry_api(endpoint: str, json: bool = True, log_widget: Any | None = None) -> dict[str, dict] | str:
    """
    Sends GET request to opentofu providers registry to a given endpoint
    and returns the response either as a JSON or as a string. It also "logs" the request.

    Local cache is used to save/retrieve API responses.
    """
    uri = f"https://api.opentofu.org/registry/docs/providers/{endpoint}"
    if cached_content := get_from_cache(endpoint):
        LOGGER.info(f"Using cached file for {endpoint} from {cached_file_path(endpoint)}")
        if log_widget is not None:
            log_widget.write(f"Cache hit [cyan]{cached_file_path(endpoint)}[/]")
        return jsonlib.loads(cached_content) if json else cached_content
    LOGGER.debug("Starting async client")
    async with httpx.AsyncClient(headers={"User-Agent": f"tofuref v{__version__}"}) as client:
        LOGGER.debug("Client started, sending request")
        try:
            r = await client.get(uri, timeout=config.http_request_timeout)
            LOGGER.debug("Request sent, response received")
        except Exception as e:
            LOGGER.error("Something went wrong", exc_info=e)
            if log_widget is not None:
                log_widget.write(f"Something went wrong: {e}")
            return ""

    if log_widget is not None:
        log_widget.write(f"GET [cyan]{endpoint}[/] [bold]{r.status_code}[/]")

    # Saving as text, because we are loading JSON if desired during cache hit
    save_to_cache(endpoint, r.text)

    return r.json() if json else r.text



from textual.widgets import RichLog

from tofuref import __version__


class CustomRichLog(RichLog):
    def __init__(self, **kwargs):
        super().__init__(id="log", markup=True, wrap=True, classes="bordered hidden", **kwargs)
        self.border_title = "Log"
        self.border_subtitle = f"tofuref v{__version__}"
        self.display = False



import json
from dataclasses import dataclass, field
from pathlib import Path
from typing import Literal

from platformdirs import user_cache_path

KIND_TYPE = Literal["resources", "providers"]


@dataclass
class Bookmarks:
    saved: dict[KIND_TYPE, list[str]] | None = None
    folder_path: Path = field(default_factory=lambda: user_cache_path("tofuref", ensure_exists=True))
    filename: str = "bookmarks.json"

    def __post_init__(self):
        self.load_from_disk()

    @property
    def path(self) -> Path:
        return self.folder_path / self.filename

    def check(self, kind: KIND_TYPE, identifier: str) -> bool:
        return identifier in self.saved[kind]

    def add(self, kind: KIND_TYPE, identifier: str):
        if not self.check(kind, identifier):
            self.saved[kind].append(identifier)
            self.save_to_disk()

    def remove(self, kind: KIND_TYPE, identifier: str):
        if self.check(kind, identifier):
            self.saved[kind].remove(identifier)
            self.save_to_disk()

    def save_to_disk(self):
        self.path.write_text(json.dumps(self.saved))

    def load_from_disk(self):
        if not self.path.exists():
            self.saved = {
                "providers": [],
                "resources": [],
            }
        else:
            self.saved = json.loads(self.path.read_text())



import re
from typing import ClassVar

from textual.binding import Binding, BindingType
from textual.widgets import MarkdownViewer, Tree

from tofuref import __version__
from tofuref.data.helpers import CODEBLOCK_REGEX


class ContentWindow(MarkdownViewer):
    ALLOW_MAXIMIZE = True

    BINDINGS: ClassVar[list[BindingType]] = [
        Binding("up", "up", "Scroll Up", show=False),
        Binding("k", "up", "Scroll Down", show=False),
        Binding("down", "down", "Scroll Down", show=False),
        Binding("j", "down", "Scroll Up", show=False),
        Binding("pageup", "page_up", "Page Up", show=False),
        Binding("ctrl+f", "page_up", "Page Up", show=False),
        Binding("pagedown", "page_down", "Page Down", show=False),
        Binding("ctrl+b", "page_down", "Page Down", show=False),
        Binding("home", "scroll_home", "Top", show=False),
        Binding("end", "scroll_end", "Bottom", show=False),
        Binding("G", "scroll_end", "Bottom", show=False),
        Binding("u", "yank", "Copy code blocks", show=False),
        Binding("y", "yank", "Copy code blocks"),
        Binding("t", "toggle_toc", "Toggle TOC"),
    ]

    def __init__(self, content=None, **kwargs):
        welcome_content = f"""
# Welcome to tofuref {__version__}!

Changelog: https://github.com/djetelina/tofuref/blob/main/CHANGELOG.md

## Controls

### Actions
| keybindings | action |
|------|--------|
| `s`, `/` | **search** in the context of providers and resources |
| `u`, `y` | Context aware copying (using a provider/resource) |
| `v` | change active provider **version** |
| `b` | persistently bookmark an item to prioritize them in sorting when next re-ordered |
| `q`, `ctrl+q` | **quit** tofuref |
| `t` | toggle **table of contents** from content window |
| `ctrl+l` | display **log** window |

### Focus windows

| keybindings | action |
|------|--------|
| `tab` | focus next window |
| `shift+tab` | focus previous window |
| `p` | focus **providers** window |
| `r` | focus **resources** window |
| `c` | focus **content** window |
| `f` | toggle **fullscreen** mode |

### Navigate in a window

Navigate with arrows/page up/page down/home/end or your mouse.

VIM keybindings should be also supported in a limited capacity.

---

# Get in touch
* GitHub: https://github.com/djetelina/tofuref"""

        self.content = content if content is not None else welcome_content
        super().__init__(
            self.content,
            classes="content",
            show_table_of_contents=False,
            id="content",
            **kwargs,
        )

    def action_up(self) -> None:
        self.document.scroll_up()

    def action_down(self) -> None:
        self.document.scroll_down()

    def action_page_down(self) -> None:
        self.document.action_page_down()

    def action_page_up(self) -> None:
        self.document.action_page_up()

    def action_scroll_home(self) -> None:
        self.document.scroll_home()

    def action_scroll_end(self) -> None:
        self.document.scroll_end()

    def update(self, markdown: str) -> None:
        self.content = markdown
        self.document.update(markdown)

    def action_toggle_toc(self):
        self.show_table_of_contents = not self.show_table_of_contents
        if not self.table_of_contents.border_title:
            self.table_of_contents.border_title = "Table of Contents"
        if self.show_table_of_contents:
            toc = self.table_of_contents.query_one(Tree)
            toc.focus()
            toc.action_cursor_down()
        else:
            self.document.focus()

    def action_yank(self):
        code_blocks = re.findall(CODEBLOCK_REGEX, self.content, re.MULTILINE | re.DOTALL)
        if self.app.code_block_selector.has_parent:
            self.app.code_block_selector.parent.remove_children([self.app.code_block_selector])
        if not code_blocks:
            return
        self.screen.mount(self.app.code_block_selector)
        self.app.code_block_selector.set_new_options(code_blocks)
        self.screen.maximize(self.app.code_block_selector)

    # Without this, the Markdown viewer would try to open a file on a disk, while the Markdown itself will open a browser link (desired)
    async def go(self, location):
        return None



from dataclasses import dataclass, field
from typing import Any

from textual.content import Content

from tofuref.config import config
from tofuref.data import emojis
from tofuref.data.bookmarks import Bookmarks
from tofuref.data.cache import cached_file_path, clear_from_cache
from tofuref.data.helpers import (
    get_registry_api,
    header_markdown_split,
)
from tofuref.data.meta import Item
from tofuref.data.resources import Resource, ResourceType


@dataclass
class Provider(Item):
    organization: str
    name: str
    description: str
    fork_count: int
    blocked: bool
    popularity: int
    _overview: str | None = None
    _active_version: str | None = None
    versions: list[dict[str, str]] = field(default_factory=list)
    fork_of: str | None = None
    raw_json: dict | None = None
    resources: list[Resource] = field(default_factory=list)
    datasources: list[Resource] = field(default_factory=list)
    functions: list[Resource] = field(default_factory=list)
    guides: list[Resource] = field(default_factory=list)
    log_widget: Any | None = None
    bookmarked: bool = False
    _cached: bool | None = None
    kind = "providers"

    @classmethod
    def from_json(cls, data: dict) -> "Provider":
        return cls(
            organization=data["addr"]["namespace"],
            name=data["addr"]["name"],
            description=data["description"],
            fork_count=data["fork_count"],
            blocked=data["is_blocked"],
            popularity=data["popularity"],
            versions=data["versions"],
            fork_of=data.get("fork_of", {}).get("display"),
            raw_json=data,
        )

    @property
    def display_name(self) -> str:
        return f"{self.organization}/{self.name}"

    @property
    def identifying_name(self) -> str:
        return self.display_name

    @property
    def active_version(self) -> str:
        if self._active_version is None:
            self._active_version = self.versions[0]["id"]
        return self._active_version

    @active_version.setter
    def active_version(self, value: str) -> None:
        self._active_version = value
        self.resources = []
        self._overview = None

    @property
    def endpoint(self) -> str:
        return f"{self.organization}/{self.name}/{self.active_version}/index.md"

    def _endpoint_wildcard_version(self) -> str:
        return self.endpoint.replace(self.active_version, "*")

    @property
    def cached(self) -> bool:
        if self._cached is None:
            return cached_file_path(self._endpoint_wildcard_version(), glob=True).exists()
        return self._cached

    @property
    def use_configuration(self) -> str:
        return f"""    {self.name} = {{
      source  = "{self.organization}/{self.name}"
      version = "{self.active_version.lstrip("v")}"
    }}"""

    async def overview(self) -> str:
        if self._overview is None:
            self._overview = await get_registry_api(self.endpoint, json=False, log_widget=self.log_widget)
            _, self._overview = header_markdown_split(self._overview)
            self._cached = True
        return self._overview

    async def load_resources(self, bookmarks: Bookmarks) -> None:
        if self.resources:
            self.sort_resources()
        await self.reload_resources(bookmarks)

    async def reload_resources(self, bookmarks: Bookmarks) -> None:
        self.resources = []
        resource_data = await get_registry_api(
            f"{self.organization}/{self.name}/{self.active_version}/index.json",
            log_widget=self.log_widget,
        )
        for g in sorted(resource_data["docs"]["guides"], key=lambda x: x["name"]):
            self.resources.append(Resource(g["name"], self, type=ResourceType.GUIDE))
        for r in sorted(resource_data["docs"]["resources"], key=lambda x: x["name"]):
            self.resources.append(Resource(r["name"], self, type=ResourceType.RESOURCE))
        for d in sorted(resource_data["docs"]["datasources"], key=lambda x: x["name"]):
            self.resources.append(Resource(d["name"], self, type=ResourceType.DATASOURCE))
        for f in sorted(resource_data["docs"]["functions"], key=lambda x: x["name"]):
            self.resources.append(Resource(f["name"], self, type=ResourceType.FUNCTION))

        for resource in self.resources:
            if bookmarks.check("resources", resource.identifying_name):
                resource.bookmarked = True

        self.sort_resources()

    def sort_resources(self) -> None:
        type_order = {ResourceType.GUIDE: 0, ResourceType.RESOURCE: 1, ResourceType.DATASOURCE: 2, ResourceType.FUNCTION: 3}

        self.resources.sort(key=lambda x: (-x.bookmarked, -x.cached, type_order[x.type], x.name))

    def visualize(self) -> Content:
        cached_icon = emojis.CACHE if config.theme.emoji else "[$success]C[/] "
        bookmark_icon = emojis.BOOKMARK if config.theme.emoji else "[$success]B[/] "
        if self.bookmarked:
            prefix = bookmark_icon
        elif self.cached:
            prefix = cached_icon
        else:
            prefix = ""
        return Content.from_markup(f"{prefix}[dim italic]{self.organization}[/]/{self.name}")

    def clear_from_cache(self) -> None:
        if self.cached:
            clear_from_cache(self._endpoint_wildcard_version())
            # Also delete overview
            clear_from_cache(self._endpoint_wildcard_version().replace(".md", ".json"))
            self._cached = False



from datetime import datetime
from pathlib import Path

from platformdirs import user_cache_path

from tofuref.config import config


def cached_file_path(endpoint: str, glob: bool = False) -> Path:
    """
    Args:
        endpoint: http endpoint of the registry API
        glob: Looks for glob matches in the cache directory, never use with saving into cache.

    Returns:
        Path to the cached file for a given endpoint.
        If glob is True, returns the first match, check `exists()`!.
    """
    filename = endpoint.replace("/", "_")
    if glob:
        matches = list(user_cache_path("tofuref", ensure_exists=True).glob(filename))
        if matches:
            return Path(matches[0])
    return user_cache_path("tofuref", ensure_exists=True) / filename


def save_to_cache(endpoint: str, contents: str) -> None:
    cached_file = cached_file_path(endpoint)
    cached_file.write_text(contents)


def is_provider_index_expired(file: Path) -> bool:
    """
    Provider index is mutable, we consider it expired after 31 days (unconfigurable for now)

    One request per month is not too bad (we could have static fallback for the cases where this is hit when offline).
    New providers that people actually want probably won't be showing too often, so a month should be okay.
    """
    timeout = config.index_cache_duration_days * 86400
    now = datetime.now().timestamp()
    return file == cached_file_path("index.json") and now - file.stat().st_mtime >= timeout


def get_from_cache(endpoint: str) -> str | None:
    cached_file = cached_file_path(endpoint)
    if not cached_file.exists() or is_provider_index_expired(cached_file):
        return None
    return cached_file.read_text()


def clear_from_cache(endpoint: str) -> None:
    cached_file = cached_file_path(endpoint, glob=True)
    while cached_file.exists():
        cached_file.unlink()
        cached_file = cached_file_path(endpoint, glob=True)



"""
Used for generating fallback/providers.json

Fetches full registry index, gets the 50 most popular providers and their 5 most recent versions
and saves that as a fallback response to the API call.

The fallback is also used for testing purposes.
"""

import json

import httpx

NUMBER_OF_PROVIDERS = 50
NUMBER_OF_VERSIONS = 5


def main():
    data = httpx.get("https://api.opentofu.org/registry/docs/providers/index.json").json()

    providers = [x for x in data["providers"] if x["addr"]["namespace"] != "terraform-providers"]

    top_providers = sorted(providers, key=lambda p: p["popularity"], reverse=True)[:NUMBER_OF_PROVIDERS]

    for provider in top_providers:
        provider["versions"] = provider["versions"][:NUMBER_OF_VERSIONS]
    with open("providers.json", "w") as f:  # noqa: PTH123
        json.dump({"providers": top_providers}, f, indent=2)


if __name__ == "__main__":
    main()



import logging
import sys
from collections.abc import Iterable
from typing import ClassVar

import httpx
from packaging.version import Version
from rich.markdown import Markdown
from textual import on
from textual.app import App, ComposeResult, SystemCommand
from textual.binding import Binding, BindingType
from textual.containers import Container
from textual.screen import Screen
from textual.theme import BUILTIN_THEMES
from textual.widgets import (
    Footer,
    Input,
    OptionList,
    Select,
)

from tofuref import __version__
from tofuref.config import config
from tofuref.data.bookmarks import Bookmarks
from tofuref.widgets import (
    CodeBlockSelect,
    ContentWindow,
    CustomRichLog,
    ProvidersOptionList,
    ResourcesOptionList,
    SearchInput,
)

LOGGER = logging.getLogger(__name__)


class TofuRefApp(App):
    CSS_PATH = "tofuref.tcss"
    BINDINGS: ClassVar[list[BindingType]] = [
        ("/", "search", "Search"),
        Binding("s", "search", "Search", show=False),
        ("y", "use", "Use provider"),
        Binding("u", "use", "Use provider", show=False),
        ("v", "version", "Provider Version"),
        Binding("p", "providers", "Providers", show=False),
        Binding("r", "resources", "Resources", show=False),
        Binding("c", "content", "Content", show=False),
        Binding("f", "fullscreen", "Fullscreen Mode"),
        Binding("ctrl+l", "log", "Show Log", show=False),
        Binding("q", "quit", "Quit"),
    ]
    TITLE = "TofuRef - OpenTofu Provider Reference"
    ESCAPE_TO_MINIMIZE = False

    def __init__(self, *args, **kwargs):
        # We are updating config in the tests, we need to reload config
        if "pytest" in sys.modules:
            config.load(reset=True)
        # We have to do this super early, otherwise tests are flaky
        for theme in BUILTIN_THEMES.values():
            theme.variables.update({"border-style": config.theme.borders_style})

        super().__init__(*args, **kwargs)
        # Widgets for easier reference, they could be replaced by query method
        self.log_widget = CustomRichLog()
        self.content_markdown = ContentWindow()
        self.navigation_providers = ProvidersOptionList()
        self.navigation_resources = ResourcesOptionList()
        self.search = SearchInput()
        self.code_block_selector = CodeBlockSelect()

        # Internal state
        self.fullscreen_mode = False
        self.providers = {}
        self.bookmarks = Bookmarks()
        self.active_provider = None
        self.active_resource = None

        self.theme = config.theme.ui

    def get_system_commands(self, screen: Screen) -> Iterable[SystemCommand]:
        yield from super().get_system_commands(screen)
        yield SystemCommand("Log", "Toggle log widget (^l)", self.action_log)

    def compose(self) -> ComposeResult:
        # Navigation
        with Container(id="sidebar"), Container(id="navigation"):
            yield self.navigation_providers
            yield self.navigation_resources

        # Main content area
        with Container(id="content"):
            yield self.content_markdown

        yield self.log_widget

        yield Footer()

    async def on_ready(self) -> None:
        LOGGER.debug("Starting on ready")

        self.content_markdown.document.code_dark_theme = config.theme.codeblocks
        self.content_markdown.document.classes = "bordered content"
        self.content_markdown.document.border_title = "Content"
        self.content_markdown.document.border_subtitle = "Welcome"

        fullscreen_threshold = config.fullscreen_init_threshold
        if self.size.width < fullscreen_threshold:
            self.fullscreen_mode = True
        if self.fullscreen_mode:
            self.navigation_providers.styles.column_span = 2
            self.navigation_resources.styles.column_span = 2
            self.content_markdown.styles.column_span = 2
            self.screen.maximize(self.navigation_providers)

        self.navigation_providers.loading = True
        self.screen.refresh()
        LOGGER.debug("Starting on ready done, running preload worker")
        self.app.run_worker(self._preload, name="preload")
        self.call_later(self.check_for_new_version)

    async def _preload(self) -> None:
        LOGGER.debug("preload start")
        self.log_widget.write("Populating providers from the registry API")
        self.providers = await self.navigation_providers.load_index()
        self.log_widget.write(f"Providers loaded ([cyan bold]{len(self.providers)}[/])")
        self.navigation_providers.populate()
        self.navigation_providers.loading = False
        self.navigation_providers.highlighted = 0
        self.log_widget.write(Markdown("---"))
        LOGGER.info("Initial load complete")

    async def check_for_new_version(self) -> None:
        newest_version = await get_current_pypi_version()
        version = Version(__version__)
        if version < newest_version:
            self.notify(f"✨ Version {newest_version} is available!\n[dim]Update now for the latest improvements[/dim]", timeout=20)

    def action_search(self) -> None:
        """Focus the search input."""
        if self.search.has_parent:
            self.search.parent.remove_children([self.search])
        for searchable in [self.navigation_providers, self.navigation_resources]:
            if searchable.has_focus:
                self.search.value = ""
                searchable.mount(self.search)
                self.search.focus()
                self.search.offset = searchable.offset + (  # noqa: RUF005
                    0,
                    searchable.size.height - 3,
                )

    async def action_use(self) -> None:
        if not self.content_markdown.document.has_focus:
            if self.active_provider:
                to_copy = self.active_provider.use_configuration
            elif self.navigation_providers.highlighted is not None:
                highlighted_provider = self.navigation_providers.options[self.navigation_providers.highlighted].prompt
                to_copy = highlighted_provider.use_configuration
            else:
                return
            self.copy_to_clipboard(to_copy)
            self.notify(to_copy, title="Copied to clipboard", timeout=10)

    def action_log(self) -> None:
        self.log_widget.display = not self.log_widget.display

    def action_providers(self) -> None:
        if self.fullscreen_mode:
            self.screen.maximize(self.navigation_providers)
        self.navigation_providers.focus()

    def action_resources(self) -> None:
        if self.fullscreen_mode:
            self.screen.maximize(self.navigation_resources)
        self.navigation_resources.focus()

    def action_content(self) -> None:
        if self.fullscreen_mode:
            self.screen.maximize(self.content_markdown)
        self.content_markdown.document.focus()

    def action_fullscreen(self) -> None:
        if self.fullscreen_mode:
            self.fullscreen_mode = False
            self.navigation_providers.styles.column_span = 1
            self.navigation_resources.styles.column_span = 1
            self.content_markdown.styles.column_span = 1
            self.screen.minimize()
        else:
            self.fullscreen_mode = True
            self.navigation_providers.styles.column_span = 2
            self.navigation_resources.styles.column_span = 2
            self.content_markdown.styles.column_span = 2
            self.screen.maximize(self.screen.focused)

    async def action_version(self) -> None:
        if self.active_provider is None:
            self.notify(
                "Provider Version can only be changed after one is selected.",
                title="No provider selected",
                severity="warning",
            )
            return
        if self.navigation_resources.children:
            await self.navigation_resources.remove_children("#version-select")
        else:
            version_select = Select.from_values(
                (v["id"] for v in self.active_provider.versions),
                prompt="Select Provider Version",
                allow_blank=False,
                value=self.active_provider.active_version,
                id="version-select",
            )
            await self.navigation_resources.mount(version_select)
            version_select.action_show_overlay()

    @on(Select.Changed, "#version-select")
    async def change_provider_version(self, event: Select.Changed) -> None:
        if event.value != self.active_provider.active_version:
            self.active_provider.active_version = event.value
            await self.navigation_resources.load_provider_resources(self.active_provider)
            await self.navigation_resources.remove_children("#version-select")

    @on(Input.Changed, "#search")
    def search_input_changed(self, event: Input.Changed) -> None:
        query = event.value.strip()
        if self.search.parent == self.navigation_providers:
            if not query:
                self.navigation_providers.populate()
            else:
                self.navigation_providers.populate([v for p, v in self.providers.items() if query in p])
        elif self.search.parent == self.navigation_resources:
            if not query:
                self.navigation_resources.populate(
                    self.active_provider,
                )
            else:
                self.navigation_resources.populate(
                    self.active_provider,
                    [r for r in self.active_provider.resources if query in r.name],
                )

    @on(Input.Submitted, "#search")
    def search_input_submitted(self, event: Input.Submitted) -> None:
        event.control.parent.focus()
        event.control.parent.highlighted = 0
        event.control.parent.remove_children([event.control])

    @on(OptionList.OptionSelected)
    async def option_list_option_selected(self, event: OptionList.OptionSelected) -> None:
        await event.control.on_option_selected(event.option)


async def get_current_pypi_version() -> Version:
    async with httpx.AsyncClient(headers={"User-Agent": f"tofuref v{__version__}"}) as client:
        try:
            r = await client.get("https://pypi.org/pypi/tofuref/json", timeout=config.http_request_timeout)
        except Exception as _:
            return Version("0.0.0")
        return Version(r.json()["info"]["version"])


def main() -> None:
    LOGGER.debug("Starting tofuref")
    TofuRefApp().run()


if __name__ == "__main__":
    main()
